import re
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

sns.set()

TITLE_SIZE = 15


def gini(x):
    # Mean absolute difference
    mad = np.abs(np.subtract.outer(x, x)).mean()
    rmad = mad / np.mean(x)
    # Gini coefficient
    return 0.5 * rmad


def lorenz(v):
    x = np.linspace(0., 100., 21)
    total = float(np.sum(v))
    y = []
    for xi in x:
        yi_vals = v[v <= np.percentile(v, xi)]
        yi = (np.sum(yi_vals) / total) * 100.0
        y.append(yi)
    return x, y


def _ts_to_string(ts):
    pattern = r"(\d+)([a-zA-Z])"
    r = re.match(pattern, ts)
    MAP = dict(W="Weeks", D="Days", M="Months", Y="Years")
    unit = MAP[r.group(2).upper()]
    count = r.group(1)
    return f"{count} {unit}"


def _in_sprint(df, timeslot="2W"):
    df_resampled = df.resample(timeslot).sum()
    df_resampled["authors"] = df["author"].resample(timeslot).nunique()
    df_resampled["addedlines"] = df_resampled["insertions"] - df_resampled["deletions"]
    return df_resampled


def _plot_gini(df, plt):
    v = df.lines.values
    bins, result = lorenz(v)
    gi = gini(v)
    plt.plot(bins, result, label="commit")
    plt.plot(bins, bins, '--', label="perfect equality")
    plt.xlabel("Percentile")
    plt.ylabel("Ratio")
    plt.xlim(0, 100)
    plt.fill_between(bins, bins, result, color="blue", alpha=0.2)
    plt.fill_between(bins, result, color="red", alpha=0.2)
    title_label = f"GINI COEFFICIENT: {gi:.1%}"
    plt.title(title_label, fontsize=TITLE_SIZE)
    plt.legend()
    pass


def _plot_hist(df, plt, ts):
    v = df.lines.values
    median = np.median(v)
    timeslot = _ts_to_string(ts)
    sns.distplot(v)
    plt.xlim(0,)

    plt.title("Histgram of Code Output", fontsize=TITLE_SIZE)
    _plot_text(plt, f"median={int(median) :,d} lines")
    plt.ylabel("")
    plt.xlabel(f"Code Output in {timeslot}")


def _plot_text(plt, text):
    ax = plt.gca()
    plt.text(0.99, 0.1, text,
             horizontalalignment='right',
             verticalalignment='top',
             bbox=dict(facecolor='#cccccc', alpha=0.5),
             transform=ax.transAxes)


def _plot_authors(df, plt):
    date = df.index.values
    authors = df["authors"]
    mean = authors.mean()
    plt.title(" Number of Actual Authors", fontsize=TITLE_SIZE)
    plt.plot(date, authors, marker=".", label="commit authors")
    plt.plot(date, np.ones(len(authors)) * mean, "--", label="mean")
    _plot_text(plt, f"mean = {mean :.1f} authors/timeslot")
    plt.xlim(date[0], date[-1])
    plt.ylabel("Unique number of committed author")
    plt.legend()


def _plot_code(df, plt):
    date = df.index.values
    total_change = df.lines.sum()
    total_added = (df.insertions - df.deletions).sum()
    refactor = 1 - total_added / total_change
    plt.title("Code Output", fontsize=TITLE_SIZE)
    plt.plot(date, df.lines, label="lines")
    plt.plot(date, df.insertions, color="g", label="insertions")
    plt.plot(date, df.deletions, color="r", label="deletions")
    _plot_text(
        plt,
        f"lines={total_change:,d} ,added={total_added:,d}, refactor={refactor:.2f}")
    plt.xlim(date[0], date[-1])
    plt.ylabel("Lines")
    plt.fill_between(
        date,
        df.insertions,
        df.deletions,
        where=df.insertions >= df.deletions,
        color="g",
        alpha=0.5,
        interpolate=True)
    plt.fill_between(
        date,
        df.insertions,
        df.deletions,
        where=df.insertions < df.deletions,
        color="r",
        alpha=0.5,
        interpolate=True)
    plt.legend()


def plot(df, timeslot='2W', output=False, name="[This Graph]"):
    suptitle = f"{name} : created by 'gilot'"

    dfs = _in_sprint(df, timeslot=timeslot)
    plt.figure(figsize=(16, 9))
    plt.tight_layout(pad=0.05, w_pad=0)
    plt.suptitle(suptitle, fontsize=13, y=0.95, x=0.8)
    plt.subplots_adjust(wspace=0.15, hspace=0.4)
    # PLOT GINI / LORENTZ
    plt.subplot(2, 2, 1)
    _plot_gini(dfs, plt)

    # PlOT HIST
    plt.subplot(2, 2, 3)
    _plot_hist(dfs, plt, timeslot)

    # PLOT CODE
    plt.subplot(2, 2, 2)
    _plot_code(dfs, plt)

    # PLOT AUTHORS
    plt.subplot(2, 2, 4)
    _plot_authors(dfs, plt)
    if(output):
        plt.savefig(output, dpi=150, bbox_inches='tight')
    else:
        plt.show()


def info(df, timeslot="2W"):
    rdf = _in_sprint(df, timeslot)
    desc = rdf.describe().drop("count")
    dic = desc.to_dict()
    sdf = df.sum()
    lines = int(sdf.lines)
    added = int(sdf.insertions - sdf.deletions)

    output = dict(lines=lines,added=added, refactor=1 - added / lines)
    # dic["lines"]["sum"] = rdf.sum().values
    return dict(
        gini=gini(rdf.lines.values),
        output=output,
        since=str(df.index.values[-1]),
        until=str(df.index.values[1]),
        timeslot=_ts_to_string(timeslot),
        **dic
    )



import json
import sys
import datetime

import git
import argparse

import gilot
from dateutil.relativedelta import relativedelta

parser = argparse.ArgumentParser(description="""
gilot is a tool for analyzing and visualizing git logs

1) simple way (1 liner using pipe)
! gilot log REPO_DIR | gilot plot

2) 2-phase way
! gilot log REPO_DIR > repo.csv
! gilot plot -i repo.csv -o graph.png


""", formatter_class=argparse.RawDescriptionHelpFormatter)


def handle_log(args):
    df = gilot.from_repo(
        args.repo,
        branch=args.branch,
        since=args.since or args.month)
    df.to_csv(args.output)


def handle_plot(args):
    df = gilot.from_csvs(args.input)
    gilot.plot(df, output=args.output, name=args.name, timeslot=args.timeslot)


def handle_info(args):
    df = gilot.from_csvs(args.input)
    print(json.dumps(gilot.info(df), indent=4, sort_keys=False))


# str -> date 型変換関数
def _type_date(date_str):
    return datetime.date.fromisoformat(date_str)


def _type_date_period(months):
    delta = relativedelta(months=int(months))
    return datetime.date.today() - delta


def _type_repo(repo_dir):
    try:
        return git.Repo(repo_dir)
    except BaseException:
        print(f"repo:{repo_dir} must be git repository\n")
        parser.print_help()
        sys.exit(-1)


subparsers = parser.add_subparsers()

# log コマンドの parser を作成
parser_log = subparsers.add_parser(
    'log', help='make git log csv data/ see `log -h`')

parser_log.add_argument(
    'repo',
    type=_type_repo,
    help='REPO must be a root dir of git repository')

parser_log.add_argument(
    "-b", "--branch",
    help="target branch name. default 'origin/HEAD' ",
    default="origin/HEAD")

parser_log.add_argument(
    "-o", "--output",
    default=sys.__stdout__)

parser_log.add_argument(
    "--since",
    help="SINCE must be ISO format like 2020-01-01.",
    type=_type_date)

parser_log.add_argument(
    "--month",
    help="MONTH is how many months of log data to output. default is 6",
    default=_type_date_period(6),
    type=_type_date_period)

parser_log.set_defaults(handler=handle_log)

# plot コマンドの parser を作成
parser_plot = subparsers.add_parser(
    'plot', help='plot graph using the csv file see `log -h`')

parser_plot.add_argument(
    '-i', "--input",
    nargs="*",
    default=[sys.__stdin__])

parser_plot.add_argument(
    '-t', "--timeslot",
    help="resample period like 2W or 7D or 1M ",
    default="2W")
parser_plot.add_argument(
    '-o', "--output",
    default=False,
    help="OUTPUT FILE ")
parser_plot.add_argument(
    "-n", "--name",
    default="GIT LOG REPORT",
    help="name")

parser_plot.set_defaults(handler=handle_plot)

# info コマンドの parser を作成
parser_info = subparsers.add_parser(
    'info', help='plot graph using the csv file see `log -h`')

parser_info.add_argument(
    '-i', "--input",
    nargs="*",
    default=[sys.__stdin__])

parser_info.add_argument(
    '-t', "--timeslot",
    help="resample period like 2W or 7D or 1M ",
    default="2W")

parser_info.set_defaults(handler=handle_info)


def main():
    # コマンドライン引数をパースして対応するハンドラ関数を実行
    args = parser.parse_args()
    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        # 未知のサブコマンドの場合はヘルプを表示
        parser.print_help()


if __name__ == "__main__":
    main()





import git
import datetime
import pandas as pd

from dateutil.relativedelta import relativedelta

DF_NULL = pd.DataFrame(
    [],
    columns=[
        "date",
        "hexsha",
        "author",
        "insertions",
        "deletions",
        "lines",
        "files"])
DF_NULL.set_index("date", inplace=True)


def dateformat(date):
    return datetime.datetime.fromtimestamp(date).strftime("%Y-%m-%d %H:%M:%S")


def commit_to_dict(commit):
    # TODO 拡張子ごとにフィルタできたりしたほうがYAMLの編集とか入らないからいいだろうか。
    return dict(
        date=dateformat(
            commit.committed_date),
        hexsha=commit.hexsha,
        author=commit.author.name,
        **commit.stats.total)


def log(
        repo,
        branch="origin/HEAD",
        since=datetime.date.today() - relativedelta(months=6)):
    return [commit_to_dict(c) for c in repo.iter_commits(branch, since=since)]


def _df_index(df):
    df.date = pd.to_datetime(df.date)
    df.set_index("date", inplace=True)
    df.sort_index()
    return df


def from_repo(r, **kwargs):
    logs = log(r, **kwargs)
    df = pd.DataFrame.from_records(logs)
    if (len(df) == 0):
        return DF_NULL
    return _df_index(df)


def from_dir(r, **kwargs):
    return from_repo(git.Repo(r), **kwargs)


def from_csv(r):
    df = pd.read_csv(r)
    return _df_index(df)


def from_csvs(iolist):
    return pd.concat([from_csv(i) for i in iolist])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gilot']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.3,<4.0.0',
 'argparse>=1.4.0,<2.0.0',
 'datetime>=4.3,<5.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.5,<2.0.0',
 'pandas>=1.0.4,<2.0.0',
 'seaborn>=0.10.1,<0.11.0']

entry_points = \
{'console_scripts': ['gilot = gilot.app:main']}

setup_kwargs = {
    'name': 'gilot',
    'version': '0.1.3',
    'description': 'a git log visual analyzer',
    'long_description': None,
    'author': 'hirokidaichi',
    'author_email': 'hirokidaichi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
